####-----------------------------------------------------------------
# Daniel Wollschlaeger <dwoll [at] psychologie.uni-kiel.de>
# 07.05.2010
#
# Sum of Squares Type I, II, III:
# the underlying model comparisons and their calculation in R
#
# for general remarks and the R output, please see
# <http://www.uni-kiel.de/psychologie/dwoll/r/ssTypes.php>
#
# for the hypotheses formulated with cell expected values, please see
# <http://www.uni-kiel.de/psychologie/dwoll/r/doc/ssTypes_hyp.pdf>
###-----------------------------------------------------------------

####-----------------------------------------------------------------
# 3x3 unbalanced design
# data from Maxwell & Delaney 2004 p339
# dependent variable Y: depression score
# factor A: type of therapy, B: degree of severity
####-----------------------------------------------------------------

P   <- 3                              # number of groups factor A
Q   <- 3                              # number of groups factor B
g11 <- c(41, 43, 50)                  # scores in group A1/B1
g12 <- c(51, 43, 53, 54, 46)          # scores in group A1/B2
g13 <- c(45, 55, 56, 60, 58, 62, 62)  # scores in group A1/B3
g21 <- c(56, 47, 45, 46, 49)          # scores in group A2/B1
g22 <- c(58, 54, 49, 61, 52, 62)      # scores in group A2/B2
g23 <- c(59, 55, 68, 63)              # scores in group A2/B3
g31 <- c(43, 56, 48, 46, 47)          # scores in group A3/B1
g32 <- c(59, 46, 58, 54)              # scores in group A3/B2
g33 <- c(55, 69, 63, 56, 62, 67)      # scores in group A3/B3
Y   <- c(g11, g12, g13, g21, g22, g23, g31, g32, g33)  # all scores
# corresponding factors
A   <- factor(rep(1:P, c(3+5+7, 5+6+4, 5+4+6)),
              labels=paste("A", 1:P, sep=""))
B   <- factor(rep(rep(1:Q, P), c(3,5,7, 5,6,4, 5,4,6)),
              labels=paste("B", 1:Q, sep=""))

# won't need these objects anymore
rm(g11, g12, g13, g21, g22, g23, g31, g32, g33)

####-----------------------------------------------------------------
# for comparison: 3x3 balanced design with random data
# all SS types lead to identical results
####-----------------------------------------------------------------

# Njk <- 5                     # equal cell size
# P   <- 3                     # number of groups factor A
# Q   <- 3                     # number of groups factor B
# Y   <- rnorm(Njk*P*Q, 0, 1)  # all scores
# # corresponding factors
# A   <- factor(rep(1:P, Njk*Q),      labels=paste("A", 1:P, sep=""))
# B   <- factor(rep(1:Q, each=Njk*P), labels=paste("B", 1:Q, sep=""))
# rm(Njk)  # won't need this anymore

####-----------------------------------------------------------------
# function getInf3x3
# get indicator variables (predictors) for factors from design matrix
# fit all relevant regression models for 3x3 between-subjects design
# output: * residual sum of squares (RSS) for each model and its df
#         * orthogonal projection on subspace as defined
#           by the design matrix of each model
####-----------------------------------------------------------------

getInf3x3 <- function() {
  X <- model.matrix(lm(Y ~ A + B + A:B))  # ANOVA design matrix

  # numerical predictors (indicator variables) from design matrix
  idA1 <- X[ , 2]   # factor A 1st indicator
  idA2 <- X[ , 3]   # factor A 2nd indicator
  idB1 <- X[ , 4]   # factor B 1st indicator
  idB2 <- X[ , 5]   # factor B 2nd indicator
  idI1 <- X[ , 6]   # interaction A:B 1st indicator
  idI2 <- X[ , 7]   # interaction A:B 2nd indicator
  idI3 <- X[ , 8]   # interaction A:B 3rd indicator
  idI4 <- X[ , 9]   # interaction A:B 4th indicator

  ####---------------------------------------------------------------
  # RSS for each regression model and its df with the help of lm()

  # all relevant regression models from lm()
  mod1   <- lm(Y ~ 1)                      # no effect
  modA   <- lm(Y ~ idA1+idA2)              # factor A
  modB   <- lm(Y ~             idB1+idB2)  # factor B
  modAB  <- lm(Y ~ idA1+idA2 + idB1+idB2)  # factors A, B
  # factor A, interaction A:B
  modAI  <- lm(Y ~ idA1+idA2             + idI1+idI2+idI3+idI4)
  # factor B, interaction A:B
  modBI  <- lm(Y ~             idB1+idB2 + idI1+idI2+idI3+idI4)
  # full model A, B, interaction A:B
  modABI <- lm(Y ~ idA1+idA2 + idB1+idB2 + idI1+idI2+idI3+idI4)

  # RSS for all models from lm(), i.e., sum((Y - fitted(<model>))^2)
  rss1   <- sum(residuals(mod1)^2)    # no effect, i.e., total SS
  # = var(Y) * (length(Y)-1) = sum((Y - mean(Y))^2)
  rssA   <- sum(residuals(modA)^2)    # factor A
  rssB   <- sum(residuals(modB)^2)    # factor B
  rssAB  <- sum(residuals(modAB)^2)   # factors A, B
  rssAI  <- sum(residuals(modAI)^2)   # factor A, A:B
  rssBI  <- sum(residuals(modBI)^2)   # factor B, A:B
  rssABI <- sum(residuals(modABI)^2)  # full model A, B, A:B

  # degrees of freedom for RSS from each model
  N     <- length(Y)  # total N
  df1   <- N - (0+1)  # no effect:            0 predictors + mean
  dfA   <- N - (2+1)  # factor A:             2 predictors + mean
  dfB   <- N - (2+1)  # factor B:             2 predictors + mean
  dfAB  <- N - (4+1)  # factors A, B:         4 predictors + mean
  dfAI  <- N - (6+1)  # factor A, A:B:        6 predictors + mean
  dfBI  <- N - (6+1)  # factor B, A:B:        6 predictors + mean
  dfABI <- N - (8+1)  # full model A, B, A:B: 8 predictors + mean

  ####---------------------------------------------------------------
  # alternatively: get RSS for each model and its df manually
  # based on geometric interpretation

  # design matrix for each model
  one  <- rep(1, nrow(X))                     # column of 1s
  X1   <- cbind(one)                          # no effect
  Xa   <- cbind(one, idA1, idA2)              # factor A
  Xb   <- cbind(one,             idB1, idB2)  # factor B
  Xab  <- cbind(one, idA1, idA2, idB1, idB2)  # factors A, B
  # factor A, interaction A:B
  Xai  <- cbind(one, idA1, idA2,             idI1, idI2, idI3, idI4)
  # factor B, interaction A:B
  Xbi  <- cbind(one,             idB1, idB2, idI1, idI2, idI3, idI4)
  # full model A, B, interaction A:B
  Xabi <- cbind(one, idA1, idA2, idB1, idB2, idI1, idI2, idI3, idI4)

  # orthogonal projections P on subspace
  # given by the design matrix of each model
  # (numerically, X %*% tcrossprod(solve(crossprod(X)), X) would be better)
  # P*y = y^hat are predictions of the model's least squares fit

  # in a balanced design, effect term estimates are:
  M       <- mean(Y)                          # grand mean
  Mjk     <- ave(Y, A, B, FUN="mean")         # cell means
  Mj      <- ave(Y, A,    FUN="mean")         # row means
  Mk      <- ave(Y,    B, FUN="mean")         # col means
  alphaJ  <- Mj - M                           # main effect A
  betaK   <- Mk - M                           # main effect B
  gammaJK <- Mjk - (M + alphaJ + betaK)       # interaction effect

  # no effect: prediction = (weighted) grand mean
  P1 <- X1  %*% solve(t(X1) %*% X1) %*% t(X1)
  # P1 %*% Y = M

  # factor A: prediction = corresponding (weighted) A-group mean
  Pa <- Xa %*% solve(t(Xa) %*% Xa) %*% t(Xa)
  # Pa %*% Y = Mj = M + alphaJ

  # factor B: prediction = corresponding (weighted) B-group mean
  Pb <- Xb %*% solve(t(Xb) %*% Xb) %*% t(Xb)
  # Pb %*% Y = Mk = M + betaK

  # factors A, B:
  Pab <- Xab %*% solve(t(Xab) %*% Xab) %*% t(Xab)
  # only the same in a balanced design:
  # Pab %*% Y = M + alphaJ + betaK

  # factor A, interaction A:B
  Pai <- Xai %*% solve(t(Xai) %*% Xai) %*% t(Xai)
  # only the same in a balanced design && with sum-to-zero contrasts:
  # Pai %*% Y = M + alphaJ + gammaJK

  # factor B, interaction A:B
  Pbi <- Xbi %*% solve(t(Xbi) %*% Xbi) %*% t(Xbi)
  # only the same in a balanced design && with sum-to-zero contrasts:
  # Pbi %*% Y = M + betaK + gammaJK

  # full model A, B, A:B: prediction = corresponding cell mean
  Pabi <- Xabi %*% solve(t(Xabi) %*% Xabi) %*% t(Xabi)
  # Pabi %*% Y = Mjk = M + alphaJ + betaK + gammaJK

  # RSS = squared length of projections onto the orthogonal
  # complement of each model's null hypothesis subspace
  # within NxN space
  # (I-P)*y = I*y - P*y = y - P*y = y - y^hat
  # differences to model prediction, i.e., residuals

  # NxN identity matrix = "projection" onto outermost space
  ID     <- diag(N)

  # no effect:
  rss1   <- c(crossprod((ID - P1)   %*% Y))
  # = sum((Y - M)^2)

  # factor A:
  rssA   <- c(crossprod((ID - Pa)   %*% Y))
  # = sum((Y - Mj)^2) = sum((Y - (M + alphaJ))^2)

  # factor B:
  rssB   <- c(crossprod((ID - Pb)   %*% Y))
  # = sum((Y - Mk)^2) = sum((Y - (M + betaK))^2)

  # factors A, B:
  rssAB  <- c(crossprod((ID - Pab)  %*% Y))
  # only the same in a balanced design:
  # = sum((Y - (M + alphaJ + betaK))^2)

  # factor A, interaction A:B:
  rssAI  <- c(crossprod((ID - Pai)  %*% Y))
  # only the same in a balanced design && with sum-to-zero contrasts:
  # = sum((Y - (M + alphaJ + gammaJK))^2)

  # factor B, interaction A:B
  rssBI  <- c(crossprod((ID - Pbi)  %*% Y))
  # only the same in a balanced design && with sum-to-zero contrasts:
  # = sum((Y - (M + betaK + gammaJK))^2)

  # full model A, B, interaction A:B:
  rssABI <- c(crossprod((ID - Pabi) %*% Y))
  # = sum((Y-Mjk)^2) = sum((Y - (M+alphaJ+betaK+gammaJK))^2)

  # get df of each RSS from dimension of subspace complementary
  # to the null hypothesis model subspace (within NxN space)
  df1   <- qr(ID - P1)$rank    # no effect
  dfA   <- qr(ID - Pa)$rank    # factor A
  dfB   <- qr(ID - Pb)$rank    # factor B
  dfAB  <- qr(ID - Pab)$rank   # factor AB
  dfAI  <- qr(ID - Pai)$rank   # factor A, A:B
  dfBI  <- qr(ID - Pbi)$rank   # factor B, A:B
  dfABI <- qr(ID - Pabi)$rank  # full model: A, B, A:B

  # return information about RSS from each model, its df,
  # and the corresponding orthogonal projections
  return(list( rss1=rss1,   rssA=rssA,    rssB=rssB, rssAB=rssAB,
              rssAI=rssAI, rssBI=rssBI, rssABI=rssABI,
                df1=df1,     dfA=dfA,      dfB=dfB,   dfAB=dfAB,
               dfAI=dfAI,   dfBI=dfBI,   dfABI=dfABI,
                 P1=P1,       Pa=Pa,        Pb=Pb,     Pab=Pab,
                Pai=Pai,     Pbi=Pbi,     Pabi=Pabi))
}

####-----------------------------------------------------------------
# indicator variables corresponding to different effects are
# correlated -> multicollinearity in multiple regression
# depends on types of contrast codes for categorical variables
####-----------------------------------------------------------------

# dummy coding
options(contrasts=c(unordered="contr.treatment", ordered="contr.poly"))

# effect coding
options(contrasts=c(unordered="contr.sum",       ordered="contr.poly"))

# get multicollinearity information
# pairwise correlations of indicator variables
model <- lm(Y ~ A + B + A:B)
round(cor(model.matrix(model)[ , -1]), 3)

# variance inflation factors
# same for sum-to-zero contrasts like contr.sum, contr.helmert
library(car)
vif(model)

# condition number
kappa(model, exact=TRUE)
# ev <- eigen(t(model.matrix(model)) %*% model.matrix(model))$values
# sqrt(max(ev) / min(ev[ev > 0]))

####-----------------------------------------------------------------
# SS type I
# * sequential model comparisons that conform to marginality principle
# * depend on order of model terms
# * individual effect SS sum to total effect SS
# * test equality of weighted marginal expected values
#   -> hypotheses take empirical cell sizes into account
# * do not depend on coding scheme for categorical variables
####-----------------------------------------------------------------

####-----------------------------------------------------------------
# coding scheme for categorical variables doesn't matter - run with
# dummy coding for unordered factors (factory default in R)
options(contrasts=c(unordered="contr.treatment", ordered="contr.poly"))

# or with effect coding for unordered factors (sum to zero)
options(contrasts=c(unordered="contr.sum",       ordered="contr.poly"))

# result from anova(): SS type I
anova(lm(Y ~ A + B + A:B))

####-----------------------------------------------------------------
# order of model terms matters
anova(lm(Y ~ B + A + A:B))

####-----------------------------------------------------------------
# sequential model comparisons for SS type I for order Y ~ A+B+A:B
# conform to marginality principle:
# higher-order terms are entered after all corresponding lower terms
#
# main effect factor A
# <constant> vs. A
# main effect factor B: same comparison for SS type I, II
#      A     vs. A + B
# interaction: same comparison for SS type I, II and III
#      A + B vs. A + B + A:B

cmpA <- anova(lm(Y ~ 1),     lm(Y ~ A))            # factor A
cmpB <- anova(lm(Y ~ A),     lm(Y ~ A + B))        # factor B
cmpI <- anova(lm(Y ~ A + B), lm(Y ~ A + B + A:B))  # interaction

cmpA[["Sum of Sq"]][2]  # SS factor A
cmpB[["Sum of Sq"]][2]  # SS factor B
cmpI[["Sum of Sq"]][2]  # SS interaction

####-----------------------------------------------------------------
# manual calculation of model comparisons for SS type I
# get information from utility function
resI <- getInf3x3()

# SS for each effect: sequential RSS-difference between model
# and next more restrictive model
(ssIa <- resI$rss1  - resI$rssA)           # factor A
(ssIb <- resI$rssA  - resI$rssAB)          # factor B
(ssIi <- resI$rssAB - resI$rssABI)         # interaction

# effect MS: RSS-difference divided by df-difference
(msIa <- ssIa / (resI$df1  - resI$dfA))    # factor A
(msIb <- ssIb / (resI$dfA  - resI$dfAB))   # factor B
(msIi <- ssIi / (resI$dfAB - resI$dfABI))  # interaction
(msE  <- resI$rssABI / resI$dfABI)         # error MS full model

# F-values for each effect: effect MS / error MS full model
msIa / msE  # factor A
msIb / msE  # factor B
msIi / msE  # interaction

####-----------------------------------------------------------------
# individual effect SS sum to total effect SS:
# sum of individual effect SS
ssIa + ssIb + ssIi

# total effect SS from model comparison:
# RSS model with 0 predictors - RSS full model
resI$rss1 - resI$rssABI

####-----------------------------------------------------------------
# SS for each effect from geometric interpretation:
# squared length of projection onto the orthogonal complement
# of null hypothesis model subspace within next bigger model subspace
crossprod((resI$P1  - resI$Pa)   %*% Y)  # factor A
crossprod((resI$Pa  - resI$Pab)  %*% Y)  # factor B
crossprod((resI$Pab - resI$Pabi) %*% Y)  # interaction

####-----------------------------------------------------------------
# SS type I test hypotheses about main effect of factor A
# (entered first) with marginal means that are calculated
# with weighted cell means
Nj <- tapply(Y, A, length)               # total row Ns
Mj <- tapply(Y, A, mean)                 # weighted row means
gM <- mean(Y)                            # weighted grand mean

# SS type I for factor A (entered first)
(ssIa <- sum(Nj * (Mj-gM)^2))
# SS type I for factor B (entered after A) and interaction SS
# cannot be easily calculated without matrix math (like SS type II)
# error SS: sum of squared Y - corresponding cell mean
(ssE <- sum((Y - ave(Y, A, B, FUN=mean))^2))

# degrees of freedom
N   <- length(Y)      # total N
dfA <- P-1            # for SS factor A
dfE <- N - (P*Q)      # for error SS

# mean squares for main effect of A and error
(msIa <- ssIa / dfA)  # for factor A
(msE  <- ssE  / dfE)  # MS error

# F-value factor A
msIa / msE

####-----------------------------------------------------------------
# SS type II
# * test model comparisons that conform to marginality principle
# * do not depend on order of terms in model
# * test hypotheses about complexely weighted cell expected values
#   -> hypotheses take empirical cell size into account
# * individual effect SS do not sum to total effect SS
# * result does not depend on coding scheme for categorical variables
####-----------------------------------------------------------------

####-----------------------------------------------------------------
# coding scheme for categorical variables doesn't matter - run with
# dummy coding for unordered factors (factory default in R)
options(contrasts=c(unordered="contr.treatment", ordered="contr.poly"))

# or with effect coding for unordered factors (sum to zero)
options(contrasts=c(unordered="contr.sum",       ordered="contr.poly"))

library(car)  # result from Anova() from package car
Anova(lm(Y ~ A + B + A:B), type="II")

####-----------------------------------------------------------------
# order of model terms doesn't matter
Anova(lm(Y ~ B + A + A:B), type="II")

####-----------------------------------------------------------------
# manual model comparisons for SS type II
# get information from utility function
resII <- getInf3x3()

# model comparisons for SS type II conform to marginality principle:
# higher-order terms are entered after all corresponding lower terms
#
# main effect factor A
#     B vs. A + B
# main effect factor B: same comparison for SS type I, II
# A     vs. A + B
# interaction: same comparison for SS type I, II and III
# A + B vs. A + B + A:B

# SS for main effects: RSS-difference between model with all main
# effects and that model except the target effect
(ssIIa <- resII$rssB  - resII$rssAB)           # factor A
(ssIIb <- resII$rssA  - resII$rssAB)           # factor B
(ssIIi <- resII$rssAB - resII$rssABI)          # interaction

# effect MS: RSS-difference divided by df-difference
(msIIa <- ssIIa / (resII$dfB  - resII$dfAB))   # factor A
(msIIb <- ssIIb / (resII$dfA  - resII$dfAB))   # factor B
(msIIi <- ssIIi / (resII$dfAB - resII$dfABI))  # interaction
(msE   <- resII$rssABI / resII$dfABI)          # error MS full model

# F-values for each effect: effect MS / error MS full model
msIIa / msE  # factor A
msIIb / msE  # factor B
msIIi / msE  # interaction

####-----------------------------------------------------------------
# individual effect SS do not sum to total effect SS:
# sum of individual effect SS
ssIIa + ssIIb + ssIIi

# total effect SS from model comparison:
# RSS model with 0 predictors - RSS full model
resII$rss1 - resII$rssABI

####-----------------------------------------------------------------
# SS for each effect from geometric interpretation:
# squared length of projection onto the orthogonal complement
# of null hypothesis model subspace within model subspace
# with all main effects (but not the interaction)
crossprod((resII$Pb  - resII$Pab)  %*% Y)  # factor A
crossprod((resII$Pa  - resII$Pab)  %*% Y)  # factor B
crossprod((resII$Pab - resII$Pabi) %*% Y)  # interaction

####-----------------------------------------------------------------
# SS type III
# * test model comparisons that violate marginality principle
# * do not depend on order of model terms
# * individual effect SS do not sum to total effect SS
# * test equality of unweighted marginal expected values
#   -> hypotheses do not take empirical cell size into account
# * correct results only with sum-to-zero contrasts
####-----------------------------------------------------------------

####-----------------------------------------------------------------
# coding scheme for categorical variables matters
# run with dummy coding -> factory default in R, wrong results
options(contrasts=c(unordered="contr.treatment", ordered="contr.poly"))

# effect coding for unordered factors (sum to zero, correct results)
options(contrasts=c(unordered="contr.sum",       ordered="contr.poly"))

library(car)  # result from Anova() from package car
Anova(lm(Y ~ A + B + A:B), type="III")

####-----------------------------------------------------------------
# order of model terms doesn't matter
Anova(lm(Y ~ B + A + A:B), type="III")

####-----------------------------------------------------------------
# model comparisons for SS type III
# comparisons for main effects violate principle of marginality:
# in the restricted model, interaction is included, but not one
# of the corresponding main effects
#
# main effect factor A
#     B + A:B vs. A + B + A:B
# main effect factor B
# A     + A:B vs. A + B + A:B
# interaction: same comparison for SS type I, II and III
# A + B       vs. A + B + A:B

# comparisons cannot be done using anova()
# due to violation of marginality principle
# drop1() drops each term in turn anyway
# sum-to-zero coding required for correct results
(dropRes <- drop1(lm(Y ~ A + B + A:B), ~ ., test="F"))

####-----------------------------------------------------------------
# manual model comparisons leading to SS type III
# get information from utility function
resIII <- getInf3x3()

# effect SS: RSS-difference between model with all
# effects and that model except the target effect
(ssIIIa <- resIII$rssBI - resIII$rssABI)           # factor A
(ssIIIb <- resIII$rssAI - resIII$rssABI)           # factor B
(ssIIIi <- resIII$rssAB - resIII$rssABI)           # interaction

# effect MS: RSS-difference divided by df-difference
(msIIIa <- ssIIIa / (resIII$dfBI - resIII$dfABI))  # factor A
(msIIIb <- ssIIIb / (resIII$dfAI - resIII$dfABI))  # factor B
(msIIIi <- ssIIIi / (resIII$dfAB - resIII$dfABI))  # interaction
(msE    <- resIII$rssABI / resIII$dfABI)           # error MS full model

# F-values for each effect: effect MS / error MS full model
msIIIa / msE  # factor A
msIIIb / msE  # factor B
msIIIi / msE  # interaction

# individual effect SS do not sum to total effect SS:
# sum of individual effect SS
ssIIIa + ssIIIb + ssIIIi

# total effect SS from model comparison:
# RSS model with 0 predictors - RSS full model
resIII$rss1 - resIII$rssABI

####-----------------------------------------------------------------
# SS for each effect from geometric interpretation:
# squared length of projection onto the orthogonal complement
# of null hypothesis model subspace within full model subspace
crossprod((resIII$Pbi - resIII$Pabi) %*% Y)  # factor A
crossprod((resIII$Pai - resIII$Pabi) %*% Y)  # factor B
crossprod((resIII$Pab - resIII$Pabi) %*% Y)  # interaction

####-----------------------------------------------------------------
# coding scheme for categorical variables matters for models
# that violate marginality principle

# orthogonal projections onto subspace given by model's design matrix
# when using dummy coding
options(contrasts=c(unordered="contr.treatment", ordered="contr.poly"))
resD  <- getInf3x3()     # projections from utility function
PaD   <- resD$Pa         # factor A
PbD   <- resD$Pb         # factor B
PabD  <- resD$Pab        # factors A, B
PaiD  <- resD$Pai        # factor A, A:B -> violates marginality
PbiD  <- resD$Pbi        # factor B, A:B -> violates marginality
PabiD <- resD$Pabi       # full model A, B, A:B

# orthogonal projections onto subspace given by model's design matrix
# when using effect coding (sum to zero)
options(contrasts=c(unordered="contr.sum",       ordered="contr.poly"))
resE  <- getInf3x3()     # projections from utility function
PaE   <- resE$Pa         # factor A
PbE   <- resE$Pb         # factor B
PabE  <- resE$Pab        # factors A, B
PaiE  <- resE$Pai        # factor A, A:B -> violates marginality
PbiE  <- resE$Pbi        # factor B, A:B -> violates marginality
PabiE <- resE$Pabi       # full model A, B, A:B

# check equality of projections from different contrast coding schemes
# -> projections from models that violate marginality are not equal
all.equal(PaD,   PaE)    # -> equal
all.equal(PbD,   PbE)    # -> equal
all.equal(PaiD,  PaiE)   # -> not equal
all.equal(PbiD,  PbiE)   # -> not equal
all.equal(PabD,  PabE)   # -> equal
all.equal(PabiD, PabiE)  # -> equal

####-----------------------------------------------------------------
# manual calculation of SS type III for main effects based
# on unweighted cell means
# equals above results when sum-to-zero contrasts were used
Njk <- tapply(Y, list(A, B), length)     # cell sizes
Mjk <- tapply(Y, list(A, B), mean)       # cell means
Mj  <- rowMeans(Mjk)                     # unweighted row means
Mk  <- colMeans(Mjk)                     # unweighted column means

# effective marginal N: harmonic mean of cell sizes
effNj <- 1 / rowMeans(1/Njk)             # harmonic row means
effNk <- 1 / colMeans(1/Njk)             # harmonic column means

# grand mean based on corresponding marginal means
# weighted with effective marginal N
gM1 <- sum(effNj*Mj) / sum(effNj)        # based on row means
gM2 <- sum(effNk*Mk) / sum(effNk)        # based on column means

# SS type III for factors A und B
(ssIIIa <- P * sum(effNj * (Mj-gM1)^2))  # factor A
(ssIIIb <- Q * sum(effNk * (Mk-gM2)^2))  # factor B
# interaction SS cannot be easily calculated without matrix math
# error SS: sum of squared Y - corresponding cell mean
(ssE    <- sum((Y - ave(Y, A, B, FUN=mean))^2))

# degrees of freedom
dfA <- P-1                # SS factor A
dfB <- Q-1                # SS factor B
dfE <- N - (P*Q)          # error SS

# MS type III for effects and error
(msIIIa <- ssIIIa / dfA)  # factor A
(msIIIb <- ssIIIb / dfB)  # factor B
(msE    <- ssE    / dfE)  # error

# F-values
msIIIa / msE              # factor A
msIIIb / msE              # factor B
